
/*
 * Copyright (c) Des Herriott 1993, 1994
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 */

/*
 * Source file audio.o - sound implementations.  Currently we have:
 *   PC Speaker audio emulation (Linux)
 *   /dev/audio emulation (Sun Sparc, Linux, NEC 4800/330)
 *   NAS (Network Audio System) emulation (Sun Sparc, SGI, NCD MCX, VoxWare)
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#ifdef PCSPKR_AUDIO
#include <asm/io.h>
#endif

#ifdef DEV_AUDIO
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
int audio_fd = -1;
#endif

#ifdef NAS_AUDIO
#include <X11/Xlib.h>
#include <audio/audiolib.h>
AuServer *aus;
AuDeviceID outputDevice;
AuBucketID theBucket;
#endif

#include "z80.h"
#include "util.h"
#include "rtcfg.h"

#ifdef DEV_AUDIO
/* Use "audio_dev" (by default /dev/audio) to emulate the speaker. */
int
audio_init(void)
{
	if ((audio_fd = open(GetCfg(audio_dev), O_WRONLY)) == -1) {
		xzx_mesg(XZX_WARN, "can't open audio device <%s> for writing");
		return -1;
	}

	return 0;
}
#endif


#ifdef PCSPKR_AUDIO
/*
 * Linux PC speaker audio implementation.  In theory, this ought to
 * work on any Unix running on a PC console, but I haven't tried it.
 *
 * For this to succeed, it must be run with root permissions.  Therefore
 * we do a setuid() and getuid() to give up root permissions once we
 * know we've succeeded.
 */
int
audio_init(void)
{
	if (ioperm(0x61, 1, 1)) {
		xzx_mesg(XZX_WARN, "couldn't get permission for the speaker port");
		return -1;
	}

	setgid(getgid()); setuid(getuid());

	return 0;
}
#endif


#ifdef NAS_AUDIO
/*
 * Use the Network Audio System to emulate the speaker.  We do this by
 * creating a server bucket containing a very short click, which is played
 * repeatedly to simulate a speaker buzz.
 *
 * NOTE: this doesn't work yet, so don't try it!  You'll probably just
 *       end up crashing your netaudio server!
 */
int
audio_init(void)
{
	int i;

	/*
	char *audio_server;

	audio_server = GetCfg(au_serv);

	if (!audio_server || !*audio_server) 
		audio_server = getenv("AUDIOSERVER");

	if (!audio_server || !*audio_server) 
		audio_server = DisplayString(dpy);

	if (!audio_server || *audio_server) 
		audio_server = NULL;

	*/
	if ((aus = AuOpenServer(NULL, 0, NULL, 0, NULL, NULL)) == 0) {
		xzx_mesg(XZX_WARN, "can't connect to NAS server <%s>", "blah");
		return -1;
	}

	for (i = 0; i < AuServerNumDevices(aus); i++)
		if ((AuDeviceKind(AuServerDevice(aus, i)) ==
			AuComponentKindPhysicalOutput) &&
			AuDeviceNumTracks(AuServerDevice(aus, i)) == 1) {
				outputDevice = AuDeviceIdentifier(AuServerDevice(aus, i));
				break;
		}
	if (outputDevice = AuNone) {
		xzx_mesg(XZX_WARN, "can't find a physical audio output device");
		AuCloseServer(aus);
		return -1;
	}

	if ((theBucket = AuSoundCreateBucketFromFile(aus, "foo.wav",
					AuAccessAllMasks, NULL, NULL)) == AuNone) {
		xzx_mesg(XZX_WARN, "can't create a server bucket");
		AuCloseServer(aus);
		return -1;
	}

	return 0;
}
#endif


void
out_byte_sound(uns8 oldval, uns8 newval)
{
	uns8 snd;

	if ((oldval & 0x10) == (newval & 0x10))
		return;

#ifdef PCSPKR_AUDIO
	snd = inb(0x61) & 0xfc;
	if (newval & 0x10)
		outb(snd | 2, 0x61);
	else
		outb(snd, 0x61);
#endif
#ifdef DEV_AUDIO
	snd = newval & 0x10 ? 0x1f : 0x00;
	write(audio_fd, &snd, 1);
#endif
#ifdef NAS_AUDIO
	AuSoundPlayFromBucket(aus, theBucket, outputDevice,
		AuFixedPointFromFraction(50, 100), NULL, NULL, 0, NULL, 
		NULL, NULL, NULL);
#endif
}
