/*
 * Copyright (c) Des Herriott 1993, 1994
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 */

/*
 * Source file joystick.c - emulate a Kempston joystick.
 * Only works on Linux with joystick kernel patches, tested
 * with versions 0.5, 0.6 & 0.7.
 */

#ifdef JOY
#include <fcntl.h>
#include <stdio.h>
#include <linux/joystick.h>

#include "rtcfg.h"
#include "util.h"

/* Kempston joystick bits */
#define KEMPST_RIGHT 0x01
#define KEMPST_LEFT 0x02
#define KEMPST_DOWN 0x04
#define KEMPST_UP 0x08
#define KEMPST_FIRE 0x10

static int jsfd;	/* joystick file descriptor */
static int joystick_ok = 0;

static int centre_x, centre_y;

int
joystick_init(void)
{
	struct JS_DATA_TYPE js_data;
	long tmpl;
	int tmp;

	if ((jsfd = open(GetCfg(js_device), O_RDONLY)) < 1) {
		xzx_mesg(XZX_WARN, "joystick initialisation failure!");
		perror("  -> reason: open");
		return -1;
	}

	tmpl = 10; 

	if (ioctl(jsfd, JS_SET_TIMELIMIT, &tmpl) == -1) {
		xzx_mesg(XZX_WARN, "joystick initialisation failure!");
		perror("  -> reason: ioctl JS_SET_TIMELIMIT");
		return -1;
	}

	xzx_mesg(XZX_INFO, "Linux joystick option selected:");
	xzx_mesg(XZX_INFO, "  Move joystick to lower right and press either button");

	while ((read(jsfd, &js_data, JS_RETURN) > 0) && js_data.buttons == 0x00)
		;

	for(tmp=0; js_data.x > 0xff; tmp++, js_data.x >>= 1)
		;
	js_data.x=tmp;
	for(tmp=0; js_data.y > 0xff; tmp++, js_data.y >>= 1)
		;
	js_data.y=tmp;

	if (ioctl(jsfd, JS_SET_CAL, &js_data) == -1) {
		xzx_mesg(XZX_WARN, "joystick calibration failure!");
		perror("  -> reason: ioctl JS_SET_CAL");
		return -1;
	}

	while ((read(jsfd, &js_data, JS_RETURN) > 0) && js_data.buttons != 0x00)
		;
	xzx_mesg(XZX_INFO, "  Now centralise the joystick and press either button");
	while ((read(jsfd, &js_data, JS_RETURN) > 0) && js_data.buttons == 0x00)
		;
	centre_x = js_data.x; centre_y = js_data.y;

#ifdef DEBUG
	xzx_mesg(XZX_INFO, "xzx: joystick initialisation successful\n");
#endif

	joystick_ok = 1;
	return 0;
}


/* Read the current joystick position.  If it's displaced by more than
 * <js_tolerance> from centre, then OR in the appropriate direction bit.
 */
int 
joystick_read(void)
{
	struct JS_DATA_TYPE js_data;
	int res = 0;

	if (!joystick_ok)
		return 0;

	if (read(jsfd, &js_data, JS_RETURN) != JS_RETURN) {
		xzx_mesg(XZX_WARN, "lost joystick connection!");
		joystick_ok = 0;
		return 0;
	}

	/* decode the joystick return values */
	if (js_data.x < centre_x - GetCfg(js_tolerance)) res |= KEMPST_LEFT;
	else if (js_data.x > centre_x + GetCfg(js_tolerance)) res |= KEMPST_RIGHT;

	if (js_data.y < centre_y - GetCfg(js_tolerance)) res |= KEMPST_UP;
	else if (js_data.y > centre_y + GetCfg(js_tolerance)) res |= KEMPST_DOWN;

	if (js_data.buttons) res |= KEMPST_FIRE;

	return res;
}
#endif
