
/*
 * Copyright (c) Des Herriott 1993, 1994
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Ian Collier, Des Herriott
 */

#ifdef LOAD_SAVE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "z80.h"
#include "mem.h"
#include "rtcfg.h"
#include "util.h"
#include "loadsave.h"
#include "snapshot.h"

extern void (*screen_refresh)(void);			/* from xspec.c */
extern char *dialog_event_loop(char *label); 	/* from xspec.c */

#define XZX_MAGIC "ZXF1"

static char xzx_magic[4];

static FILE *in=0, *out=0;

static struct header {
   uns8 type;
   char name[10];
   uns8 length[2];
   uns8 start[2];
   uns8 line[2];
}  prepared={3,"xxxxxxxxxx",{0,0},{0,128},{0,0}};


static void
get_hdr_from_memory(struct header *h, uns16 addr)
{
	int i;

	h->type = read_byte(addr);

	for (i = 0; i < 10; i++)
		h->name[i] = read_byte(addr + i + 1);

	h->length[0] = read_byte(addr + 11);  h->length[1] = read_byte(addr + 12);
	h->start[0]  = read_byte(addr + 13);  h->start[1]  = read_byte(addr + 14);
	h->line[0]  = read_byte(addr + 15);  h->line[1]  = read_byte(addr + 16);
}


static void
put_hdr_to_memory(struct header *h, uns16 addr)
{
	int i;

	write_byte(addr, h->type);

	for (i = 0; i < 10; i++)
		write_byte(addr + 1 + i, h->name[i]);

	write_byte(addr + 11, h->length[0]); write_byte(addr + 12, h->length[1]);
	write_byte(addr + 13, h->start[0]); write_byte(addr + 14, h->start[1]);
	write_byte(addr + 15, h->line[0]); write_byte(addr + 16, h->line[1]);
}


/* Add pseudo-ops to the ROMs to force loader()/saver() to be called */
void 
set_loadsave_traps(void)
{
	int basic_rom;
	uns16 load_addr, save_addr;
	uns8 load_byte, save_byte;

	basic_rom = GET_BASIC_ROM(GetCfg(emode));

	load_addr = LOAD_ADDR;
	save_addr = SAVE_ADDR;
	/* on the +3, the SA-BYTES entry point is 4 bytes further on */
	if (GetCfg(emode) == 3) save_addr += 4;

	/* you may want to modify this switch if you use a different ROM */
	switch(GetCfg(emode)) {
		case 48:  load_byte = 0x14; save_byte = 0x21; break;
		case 128: load_byte = 0x14; save_byte = 0x21; break;
		case 3:   load_byte = 0x14; save_byte = 0x21; break;
		default:
			xzx_mesg(XZX_ERR, "impossible emulation mode?");
	}

	if (realMem[basic_rom][load_addr] != load_byte) {
		xzx_mesg(XZX_WARN, "modified ROM, LOAD disabled (got 0%02xh)", 
			realMem[basic_rom][load_addr]);
	} else {
		realMem[basic_rom][load_addr]   = 0xED;
		realMem[basic_rom][load_addr+1] = 0xFC;
	}
	if (realMem[basic_rom][save_addr] != save_byte) {
		xzx_mesg(XZX_WARN, "modified ROM, SAVE disabled (got 0%02xh)",
			realMem[basic_rom][save_addr]);
	} else {
		realMem[basic_rom][save_addr]   = 0xED;
		realMem[basic_rom][save_addr+1] = 0xFD;
	}
}


/*
 * Load a file from disk.  This is called whenever the trap instructions
 * ED FC is executed.
 */
void
loader(void)
{ 
   screen_refresh();

   if (*a == 0) {  /* header */
      /* two header structures used by BASIC; one at IX and one at IX-17. */
      struct header hd1, hd2;
      char hdrname[11];
      char *name = hdrname;
      int basic, i;

	  get_hdr_from_memory(&hd1, *ix - 17);
	  get_hdr_from_memory(&hd2, *ix);

      basic = (hd1.type < 4) && (*ix < read_word(23653));

	  /* extract the filename and remove any trailing spaces */
      memcpy(name, hd1.name, 10);
      for (i=9; i>=0&&name[i]==' '; i--)
		  ;
      name[i+1] = 0;
      if (i < 0 || name[0] == (char)255)
	  	basic = 0; /* ignore basic header for LOAD"" */

      if (in) { /* second successive header */
         fclose(in), in=0;
         name = dialog_event_loop("Header rejected - enter another file name");
gotname:
		 if (!name[0]) {   /* request was rejected */
            *pc=0x0806;  /* "tape loading error" */
            return;
         }
      } else if (!basic) {
         name = dialog_event_loop("Enter file name for load");
         if (!name[0]) {
            *pc=0x0806;
            return;
         }
      }
      in = open_xzx_file(name,"r");

      if (!in) {
		 xzx_mesg(XZX_WARN, "can't open %s", name);
         name = dialog_event_loop("Enter another file name");
         goto gotname;
      }
      if (fread(xzx_magic, 1, sizeof xzx_magic, in) < sizeof xzx_magic ||
			 fread((char*)&hd2, 1, 17, in) < 17 ||
			 memcmp(xzx_magic, XZX_MAGIC, sizeof xzx_magic)) {
         /* it isn't an xzx file.  Fill in the header */

         struct stat st;
         rewind(in);
         memcpy((char*)&hd2, (char*)&prepared, 17);
         if (basic)
			 memcpy(hd2.name, hd1.name, 10);
         else {
            for(i=0; i<10 && (hd2.name[i]=name[i]); i++)
				;
            while (i<10)
				hd2.name[i++]=' ';
         }
         if (fstat(fileno(in),&st)) {
			xzx_mesg(XZX_ERR, "fstat failed on %s", name);
            st.st_size=0;
         }
         hd2.length[0]=st.st_size%256;
         hd2.length[1]=st.st_size/256;
		 put_hdr_to_memory(&hd2, *ix);
         if (st.st_size==49179) {  /* it's a snapshot */
            fclose(in);
            in=0;
            read_sna_fmt(name);
            return;
         } else if (st.st_size>49152) {
            xzx_mesg(XZX_ERR,"%s has length %d!\n", name, st.st_size);
			return;
		 }
      } else { /* It is an xzx file.  Kludge the file name. */
         if (basic) {
		 	memcpy(hd2.name, hd1.name, 10);
         } else {
            for(i=0; i<10 && (hd2.name[i]=name[i]); i++)
				;
            while (i < 10)
				hd2.name[i++]=' ';
         }
		 put_hdr_to_memory(&hd2, *ix);
      }
   } else { /* A != 0 => data */ 
      int load = Tst(carryFlag);
      if (!in) {  /* headerless load */
         char *name = dialog_event_loop("Enter file name for headerless load");
         if (!name[0]) {
            *pc=0x0806;
            return;
         }
         in = open_xzx_file(name,"r");

		 /* If we're in +3 mode, check to see if the name was prefixed by
		  * "t:" (tape select), if so, try the filename without this prefix */
		 if (!in && GetCfg(emode) == 3 && strncasecmp(name, "t:", 2) == 0)
		   in = open_xzx_file(name + 2, "r");

         if (!in) {
			xzx_mesg(XZX_WARN, "can't open %s", name);
            Clr(carryFlag);
            ret();
            return;
         }
      }
      Set(carryFlag);
      while ((*de)--) {
         int ch=getc(in);
         if (ch==EOF) {
            Clr(carryFlag); /* carry reset on error */
            break;
         }
         else if (load) write_byte((*ix)++,ch);
         else if (read_byte((*ix)++)!=(uns8)ch) {
            Clr(carryFlag);
            break;
         }
      }
      fclose(in);
      in=0;
   }
   ret();
}


/* whenever PC gets to 04C2 */
void
saver(void)
{ 
   int i;

   screen_refresh();
   if (*a==0) { /* header */
      struct header hd;
      char hdrname[11];
      char newname[14];
      char *name=hdrname;

	  get_hdr_from_memory(&hd, *ix);

      if (out)fclose(out);/* second successive header.  Not much we can do...*/
      memcpy(name,hd.name,10);
      for(i=9;i>0&&name[i]==' ';i--);
      name[i+1]=0;
      strcpy(newname,name);
      strcat(newname,".bak");
      rename(name,newname);
      out=fopen(name,"w");
      if (!out) {
         fputs("xzx: ",stderr);
         perror(name);
         *pc=0x0806;
         return;
      }
      fwrite(XZX_MAGIC,1,sizeof xzx_magic,out);
   } else if (!out) {        /* headerless save */
      char *name = dialog_event_loop("Enter file name for headerless save");
      if (!name[0]) {
         *pc=0x0806;
         return;
      }
      out=fopen(name,"w");
      if (!out) {
         fputs("xzx: ",stderr);
         perror(name);
         *pc=0x0806;
         return;
      }
   }
   for (i = 0; i < *de; i++)
	  fputc(read_byte(*ix + i), out);

   *ix += *de;
   *de = 0;
   if (*a!=0)
	   fclose(out),out=0;
   ret();
}

#endif /* LOAD_SAVE */
