
/*
 * Copyright (c) Des Herriott 1993, 1994
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 */

/* Source file main.c - initial setup etc. */

#define Z80_MAIN

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#include "z80.h"
#include "mem.h"
#include "xspec.h"
#include "rtcfg.h"
#include "util.h"
#include "snapshot.h"

static void init_cpu(void);
static void got_sig(int signo);

void init_pages(void);
void read_roms(void);
void reset_cpu(void);

extern void insert_disk(char *dname, char drive);	/* from fdc.c */
extern void xzx_main_loop(void);					/* from emul.c */
extern void (*screen_refresh)(void);				/* from xspec.c */
#ifdef XZX_IF1
extern void init_if1(void);							/* from if1.c */
#endif

int
main(int argc, char **argv)
{
	int i;

	init_cpu();

	reset_cpu();

	/* Make sure we can't be killed off while initialising stuff, 
	 * or we might end up leaving shared memory segments lying about.  */
	signal(SIGINT,  SIG_IGN);
	signal(SIGHUP,  SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);

	start_emulator(&argc, argv);

	init_memory();
	init_pages();
	read_roms();

#ifdef XZX_IF1
	init_if1();
#endif

	srand(getpid());
	for (i = 16384; i < 23296; i++)
		write_byte(i, rand() & 0xff);
	screen_refresh();

	insert_disk(GetCfg(disk_a), 'A');
	insert_disk(GetCfg(disk_b), 'B');

	if (argc == 2) {
		read_snapshot(argv[1]);
		screen_refresh();
	}

	signal(SIGINT,  got_sig);
	signal(SIGHUP,  got_sig);
	signal(SIGTERM, got_sig);
	signal(SIGQUIT, got_sig);

	xzx_main_loop();

	return 0;
}


/* Read the ROM image <image> into real memory page <page> */
static int
load_rom_image(char *image, int page)
{
	FILE *ifp;
	size_t nread;

	if ((ifp = open_xzx_file(image, "r")) == NULL) {
		xzx_mesg(XZX_ERR, "can't open ROM image <%s>", image);
		return -1;
	}

	nread = fread(realMem[page], 1, PAGESIZE, ifp);

	/* give a warning if image is not 16K or 8K */
	if (nread != PAGESIZE && nread != PAGESIZE/2) {
		xzx_mesg(XZX_WARN, "ROM image <%s> is only %d bytes", image, nread);
	}

	fclose(ifp);

	return 0;
}


/* Set up the initial memory pages; on all machines, we page in ROM0 at
 * 0x0000, RAM5 at 0x4000, RAM2 at 0x8000 and RAM0 at 0xc000.
 */
void
init_pages(void)
{
	page_in(0, ROM0); page_in(1, RAM5);
	page_in(2, RAM2); page_in(3, RAM0);
}


/* Read the required ROM images into memory;
 * 48K ROM lives in ROM0.
 * 128K ROMs live in ROM0 & ROM1.
 * +3 ROMs live in ROM0, ROM1, ROM2 & ROM3.
 * Interface 1 ROM lives in ROM2 (but not on the +3).
 */
void
read_roms(void)
{
	if (GetCfg(emode) == 48)
		load_rom_image(GetCfg(rom48), ROM0);

	if (GetCfg(emode) == 128) {
		load_rom_image(GetCfg(rom128_0), ROM0);
		load_rom_image(GetCfg(rom128_1), ROM1);
	}

	if (GetCfg(emode) == 3) {
		load_rom_image(GetCfg(rompl3_0), ROM0);
		load_rom_image(GetCfg(rompl3_1), ROM1);
		load_rom_image(GetCfg(rompl3_2), ROM2);
		load_rom_image(GetCfg(rompl3_3), ROM3);
	}

#ifdef LOAD_SAVE
	set_loadsave_traps();
#endif

#ifdef XZX_IF1
	/* No support for Interface 1 on the +3 */
	if (GetCfg(if1_active) && GetCfg(emode) != 3)
		if (load_rom_image(GetCfg(if1_rom), ROM2) == -1) {
			xzx_mesg(XZX_WARN, "IF1 emulation disabled");
			SetCfg(if1_active, 0);
		}
#endif
}


static void
got_sig(int signo)
{
	Quit(0);
}


static void
init_cpu(void)
{
	af = &(theProcessor.af_pair);
	hl = &(theProcessor.hl_pair);
	bc = &(theProcessor.bc_pair);
	de = &(theProcessor.de_pair);
	sp = &(theProcessor.sp_reg);
	pc = &(theProcessor.pc_reg);
	ix = &(theProcessor.ix_reg);
	iy = &(theProcessor.iy_reg);

#ifdef LITTLE_ENDIAN
	a  = (uns8 *)&(theProcessor.af_pair) + 1;
	f  = (uns8 *)&(theProcessor.af_pair);
	h  = (uns8 *)&(theProcessor.hl_pair) + 1;
	l  = (uns8 *)&(theProcessor.hl_pair);
	b  = (uns8 *)&(theProcessor.bc_pair) + 1;
	c  = (uns8 *)&(theProcessor.bc_pair);
	d  = (uns8 *)&(theProcessor.de_pair) + 1;
	e  = (uns8 *)&(theProcessor.de_pair);
	pc_high = (uns8 *)&(theProcessor.pc_reg) + 1;
	pc_low  = (uns8 *)&(theProcessor.pc_reg);
#else
	a  = (uns8 *)&(theProcessor.af_pair);
	f  = (uns8 *)&(theProcessor.af_pair) + 1;
	h  = (uns8 *)&(theProcessor.hl_pair);
	l  = (uns8 *)&(theProcessor.hl_pair) + 1;
	b  = (uns8 *)&(theProcessor.bc_pair);
	c  = (uns8 *)&(theProcessor.bc_pair) + 1;
	d  = (uns8 *)&(theProcessor.de_pair);
	e  = (uns8 *)&(theProcessor.de_pair) + 1;
	pc_high = (uns8 *)&(theProcessor.pc_reg);
	pc_low  = (uns8 *)&(theProcessor.pc_reg) + 1;
#endif

	carryFlag =  addsubFlag = par_overFlag = 0;
	hcarryFlag = zeroFlag =   signFlag = 0;
}


void
reset_cpu(void)
{
	*pc = 0;

	theProcessor.int_vec = 0;
	theProcessor.refresh = 0;
	theProcessor.im = 0;
	theProcessor.iff1 = 0;
	theProcessor.iff2 = 0;

#ifdef LEVEL_LOADER
	*last_snapshot_name = 0;
#endif

	init_pages();
}


