
/*
 * Copyright (c) Des Herriott 1993, 1994
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 */

/*
 * Source file util.c - contain various common utility functions
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <ctype.h>

#include "z80.h"
#include "mem.h"
#include "util.h"
#include "rtcfg.h"


/* list of functions to be called when shutting down */
static struct on_quit_rec {
	void (*f)();
	struct on_quit_rec *next;
} *on_quit_list = 0;


/* Open a file, checking each element in the colon-separated resource
 * 'lib_dir', unless the filename is an absolute path or relative to the
 * current or parent directory.  Behaves like fopen() in other respects.
 */
FILE *
open_xzx_file(char *name, char *amode)
{
	FILE *fp = 0;
	char *buf;
	char *path, *element;

	if (!name || !name[0])
		return NULL;

	if (name[0] == '/' || strncmp(name, "./", 2) == 0 ||
			strncmp(name, "../", 3) == 0)
		return(fopen(name, amode));

	path = (char *) malloc(strlen(GetCfg(lib_dir)) + 1);
	strcpy(path, GetCfg(lib_dir));

	element = strtok(path, ":");
	while (element) {
		buf = malloc(strlen(name) + strlen(element) + 2);
		if (!buf) return(NULL);
		sprintf(buf, "%s/%s", element, name);
		fp = fopen(buf, amode);
		free(buf);
		if (fp)
			break;
		element = strtok(NULL, ":");
	}
	free(path);

	return(fp);
}


void
OnQuit(void (*fn)())
{
    struct on_quit_rec *p;

	p = (struct on_quit_rec *)malloc(sizeof (struct on_quit_rec));
	if (p) {
		p->f = fn;
		p->next = on_quit_list;
		on_quit_list = p;
	} else {
		xzx_mesg(XZX_FATAL, "malloc failed in OnQuit()!");
		exit(1);
	}

}


/* Shut down the emulator cleanly, calling any functions on the Quit
 * list.  Currently only xspec.c:closedown().
 */
void
Quit(int status)
{
	struct on_quit_rec *p = on_quit_list;

	while (p) {
		p->f();
		p = p->next;
	}

	exit(status);
}


/* Report an info, warning, error or fatal message.  Exit if a
 * fatal message has been reported.
 */
#ifdef nec_ews
/* NEC's C compiler is broken - we have to use the old varargs method */
void 
xzx_mesg(va_alist)
va_dcl
{
	va_list args;
	int error_level;
	char *fmt;

	va_start(args);
	error_level = (int)va_arg(args, int);
	fmt = (char *)va_arg(args, char *);
#else
void
xzx_mesg(int error_level, ...)
{
	va_list args;
	char *fmt;

	va_start(args, error_level);
	fmt = (char *)va_arg(args, char *);
#endif
	
	fprintf(stderr, "xzx: ");
	if (error_level == XZX_WARN)
		fprintf(stderr, "warning: ");
	else if (error_level == XZX_ERR)
		fprintf(stderr, "error: ");
	else if (error_level == XZX_FATAL)
		fprintf(stderr, "fatal: ");

	vfprintf(stderr, fmt, args);
	fputc('\n', stderr);

	if (error_level == XZX_FATAL)
		Quit(1);
}


/* Set the interrupt frequency in microseconds.  0 disables. */
/* Make sure a signal handler is set up before you use this! */
void
int_frequency(int ms)
{
	struct itimerval itv;

	if (GetCfg(icnt)) return;

	itv.it_interval.tv_sec = 0;
	itv.it_interval.tv_usec = ms;
	itv.it_value.tv_sec = 0;
	itv.it_value.tv_usec = ms;
	setitimer(ITIMER_REAL, &itv, NULL);
}


/* Implementations of strcasecmp and strncasecmp, if you don't have them */
#ifdef NEED_STRCASECMP
int strcasecmp(char *s1, char *s2)
{
	char c1, c2;

	for ( ; *s1 && *s2; s1++, s2++) {
		c1 = isupper(*s1) ? tolower(*s1) : *s1;
		c2 = isupper(*s2) ? tolower(*s2) : *s2;
		if (c1 != c2) return (c1 - c2);
	}
	if (*s1 || *s2) return (*s1 - *s2);
	return 0;
}

int strncasecmp(char *s1, char *s2, int n)
{
	int i;
	char c1, c2;

	for (i = 0; *s1 && *s2 && i < n; s1++, s2++, i++) {
		c1 = isupper(*s1) ? tolower(*s1) : *s1;
		c2 = isupper(*s2) ? tolower(*s2) : *s2;
		if (c1 != c2) return (c1 - c2);
	}
	if ((i < n) && (*s1 || *s2)) return (*s1 - *s2);
	return 0;
}
#endif

