
/*
 * Copyright (c) Des Herriott 1993, 1994
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 *  (and Thomas Ahn Kjaer for the banked port handling)
 */

/* Source file xspec.c - emulation of ZX Spectrum hardware */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#if defined(nec_ews) && !defined(__STDC__)
void *shmat(int, void*, int);	/* shut the compiler up */
#endif
#endif

#include <stdio.h>
#include <stdlib.h>

#include "z80.h"
#include "mem.h"
#include "xspec.h"
#include "xspecP.h"
#include "rtcfg.h"
#include "util.h"
#include "version.h"
#include "snapshot.h"

#include "xzx.icon"

static Display *dpy;
static Screen *sptr;
static int scrn;
static Window root;
static Colormap cmap;
static GC theGc;
static GC fgGc, bgGc;
static Window borderWin, mainWin;
static KeySym *key_table;
static XImage *xim;
static Pixmap thePix;
static XFontStruct *theFont;
#ifdef MITSHM
static XShmSegmentInfo xshminfo;
#endif

void (*screen_write)(uns16, uns8);
void (*screen_refresh)(void);

static uns8 keyports[9];	/* 8 rows of keys (+ 1 row for unscanned keys) */
static uns8 inPorts[256], outPorts[256];

static int xpos, ypos;

/* All to do with screen painting */
enum {XVal, YVal};
extern unsigned char dtbl[6144][2];		/* maps offset to (x,y) coords */
static unsigned long pixels[NCOLORS];	/* 16 Spectrum colours */
static int screen_dirty;				/* does the screen need refreshing? */
static uns16 DFILE;						/* address of base of display file */
static int x_min, x_max, y_min, y_max;	/* bounding box of last area changed */
static int bytes_per_line;				/* no. of bytes in 1 line of XImage */
static char *xim_data_ptr;				/* pointer to actual XImage data */
/* These macros map a byte in the attribute area to an X11 pixel value */
#define Background(byte) (((byte) & 0x38) >> 3) | (((byte) & 0x40) >> 3)
#define Foreground(byte) ((byte) & 0x07) | (((byte) & 0x40) >> 3)
/* Map an (x,y) pixel coordinate to a offset into the attribute area */
#define XYToAttr(x,y) (((x) >> 3) + ((y & 0xf8) << 2))

/* right-hand edge of the bounding box must be rounded up */
static uns8 round_up_x[] = { 0x07, 0x0f, 0x17, 0x1f };

static int flash_phase = 0;	/* gets toggled between 1 and 0 twice per second */
static int unshift_cursor = 0;	/* do cursor keys also press Caps Shift ? */
static int want_colour;			/* did we initially want colour ? */
int last_border_col = 7;		/* dull white border */
#if defined(DEV_AUDIO) || defined(PCSPKR_AUDIO) || defined(NAS_AUDIO)
static int sound_on;
#endif

/* from resource.c */
extern Display *xzx_open_display(int *argcp, char **argvp, KeySym *key_table);

/* all to do with the popup dialog */
#define EF_MAX_VISIBLE 50
#define EF_MAX 256
#define BUTTON_LABEL "OK"
#define GUTTER_HT 5
#define GUTTER_WID 5
static void make_dialog(int *argcp, char **argvp);
char *dialog_event_loop(char *label);
static Window dlg_win, entry_win, ok_win;
static int dlg_wid, dlg_ht, dlg_x, dlg_y;
static int char_width, font_ht;
static int label_wid, label_ht, entry_wid, entry_ht, ok_wid, ok_ht;
static int cursor_pos = 0, cur_win_pos = 0;
static char dlg_string[EF_MAX];


/* For 128K and +3, remember the last values written to the bankm & bank678
 * output ports.  Also, the screen selection flag; either 5 or 7. */
uns8 last_0x7ffd, last_0x1ffd, last_0xfffd;
int screen_select = RAM5;

/*
 *  Private functions start here
 */


#if SCALING > 1
/* Initialise the halftoning stipples for monochrome display */
static void
init_scaleup_table(void)
{
	int j, k, l, m;
	unsigned long bigval; /* SCALING must be <= 4! */

	scaleup_done = 1;

	for(l = 0; l < 256; scaleup[l++] = bigval)
		for (m = l, bigval = j = 0; j < 8; j++) {
			for(k = 0; k < SCALING; k++)
				if (BitmapBitOrder(dpy) == LSBFirst)
					bigval = (bigval << 1) | (m & 1);
				else
					bigval = (bigval >>1 ) | ((m & 1) << 31);
			m>>=1;
		}
		if ((SCALING < 4) && (ImageByteOrder(dpy) == MSBFirst))
			if (BitmapBitOrder(dpy) == MSBFirst)
				bigval >>= 32 - 8 * SCALING;
			else
				bigval <<= 32 - 8 * SCALING;

	/* reverse the shading patterns if white==1 */
	if (WhitePixel(dpy, scrn))
		for (j = 0; j < 16; j++)
			for(k = 0; k < 4; k++)
				halftone[j][k]^=255;
}
#endif


/* Screen write functions - one for XImages, one for Pixmaps */
void
screen_write_xim(uns16 addr, uns8 val)
{
	int x, y;
	uns8 *baseptr;
	uns16 offset;
	unsigned int ink, paper;
	unsigned long i_pix, p_pix, tmp;
	uns8 attr;

#if SCALING > 1 
	uns8 pmask, ipmask;
	int j, k;
	long bigval;
#endif

	screen_dirty++;

	offset = addr & 0x3fff;

	x = dtbl[offset][XVal]; y = dtbl[offset][YVal];
	attr = realMem[screen_select][6144 + XYToAttr(x,y)];
	ink = Foreground(attr); paper = Background(attr);
	/* toggle the colours if we're in the inverse part of the FLASH cycle */
	if (flash_phase && (attr & 0x80)) {
		tmp=ink; ink=paper; paper=tmp;
	}

	/* In monochrome mode, make sure that we get two contrasting colours */
	if (GetCfg(monochrome)) {
		if (paper > ink) {
			p_pix = WhitePixel(dpy, scrn); i_pix = BlackPixel(dpy, scrn);
		} else if (paper < ink) {
			i_pix = WhitePixel(dpy, scrn); p_pix = BlackPixel(dpy, scrn);
		} else {
			if (paper < 4 || (paper > 7 && paper < 12)  ) {
				i_pix = BlackPixel(dpy, scrn); p_pix = BlackPixel(dpy, scrn);
			} else {
				i_pix = WhitePixel(dpy, scrn); p_pix = WhitePixel(dpy, scrn);
			}
		}
#if SCALING>1
		pmask=255*p_pix;
		ipmask=(255*i_pix)^pmask;
#endif
	} else {
		i_pix = pixels[ink];
		p_pix = pixels[paper];
	}

/* #ifdef this just in case the compiler is braindead */
#if SCALING > 1
	x *= SCALING; y *= SCALING;
#endif

	/* Extend the bounding box if necessary */
	if (x < x_min) x_min = x;
	if (x > x_max) x_max = x + SCALING - 1;
	if (y < y_min) y_min = y;
	if (y > y_max) y_max = y + SCALING - 1;

#if SCALING == 1
	if (bytes_per_line == 256) {
		baseptr = (uns8 *)xim_data_ptr + x + (y << 8);
		baseptr[0] = val & 0x80 ? i_pix : p_pix;
		baseptr[1] = val & 0x40 ? i_pix : p_pix;
		baseptr[2] = val & 0x20 ? i_pix : p_pix;
		baseptr[3] = val & 0x10 ? i_pix : p_pix;
		baseptr[4] = val & 0x08 ? i_pix : p_pix;
		baseptr[5] = val & 0x04 ? i_pix : p_pix;
		baseptr[6] = val & 0x02 ? i_pix : p_pix;
		baseptr[7] = val & 0x01 ? i_pix : p_pix;
	} else /* bytes_per_line == 32 */ {
		baseptr = (uns8 *)xim_data_ptr + (x >> 3) + (y << 5);
		baseptr[0] = (i_pix ? val : 0) | (p_pix ? val^255 : 0);
	}
#else /* SCALING > 1 */
	if (bytes_per_line == 256 * SCALING) {
		for (j = 0; j < SCALING; j++)
			for (k = 0; k < SCALING; k++) {
				baseptr = (uns8 *)xim_data_ptr + x + y*HSize;
				baseptr[0 * SCALING + j * HSize + k] = val & 0x80? i_pix: p_pix;
				baseptr[1 * SCALING + j * HSize + k] = val & 0x40? i_pix: p_pix;
				baseptr[2 * SCALING + j * HSize + k] = val & 0x20? i_pix: p_pix;
				baseptr[3 * SCALING + j * HSize + k] = val & 0x10? i_pix: p_pix;
				baseptr[4 * SCALING + j * HSize + k] = val & 0x08? i_pix: p_pix;
				baseptr[5 * SCALING + j * HSize + k] = val & 0x04? i_pix: p_pix;
				baseptr[6 * SCALING + j * HSize + k] = val & 0x02? i_pix: p_pix;
				baseptr[7 * SCALING + j * HSize + k] = val & 0x01? i_pix: p_pix;
			}
	} else /* bytes_per_line == 32*SCALING */ {
		if (!scaleup_done)
			init_scaleup_table();

		bigval = scaleup[val];
		baseptr = (uns8 *)xim_data_ptr + (x >> 3) + (y<<5) * SCALING;
		y &= 0x03; /* y is now the line within the 8x4 shading bitmap */
		for (j = 0; j < SCALING; j++) {
			if (want_colour) {
				pmask  = halftone[paper][y];
				ipmask = halftone[ink][y]^pmask;
			}
			/* This should make it work for any client/server endian-ness */
			if (ImageByteOrder(dpy) == BitmapBitOrder(dpy)) {
				baseptr[0] = shade(bigval, ipmask, pmask);
				baseptr[1] = shade(bigval >> 8, ipmask, pmask);
#if SCALING>2
				baseptr[2] = shade(bigval >> 16, ipmask, pmask);
#endif
#if SCALING>3
				baseptr[3] = shade(bigval >> 24, ipmask, pmask);
#endif
			} else {
				baseptr[0] = shade(bigval >> 24, ipmask, pmask);
				baseptr[1] = shade(bigval >> 16, ipmask, pmask);
#if SCALING>2
				baseptr[2] = shade(bigval >> 8, ipmask, pmask);
#endif
#if SCALING>3
				baseptr[3] = shade(bigval, ipmask, pmask);
#endif
			}
			baseptr += SCALING * 32;
			if (++y == 4) y=0;
		}
	}
#endif
}

void
screen_write_pix(uns16 addr, uns8 val)
{
	int offset;
	int x, y;
	unsigned int ink, paper;
	unsigned long i_pix, p_pix, tmp;
	uns8 attr;
	static unsigned long last_i_pix;
	static unsigned long last_p_pix;
	
	int nfg = 0, nbg = 0;
	int i;
	static XPoint fg_pts[8], bg_pts[8];	/* up to 8 fg & bg pixels */

	screen_dirty++;

	offset = addr & 0x3fff;

	x = dtbl[offset][XVal]; y = dtbl[offset][YVal];
	attr = realMem[screen_select][6144 + XYToAttr(x,y)];
	ink = Foreground(attr); paper = Background(attr);
	/* toggle the colours if we're in the inverse part of the FLASH cycle */
	if (flash_phase && (attr & 0x80)) {
		tmp = ink; ink = paper; paper = tmp;
	}

	/* In monochrome mode, make sure that we get two contrasting colours */
	if (GetCfg(monochrome)) {
		if (paper > ink) {
			p_pix = WhitePixel(dpy, scrn); i_pix = BlackPixel(dpy, scrn);
		} else if (paper < ink) {
			i_pix = WhitePixel(dpy, scrn); p_pix = BlackPixel(dpy, scrn);
		} else {
			if (paper < 4 || (paper > 7 && paper < 12)  ) {
				i_pix = BlackPixel(dpy, scrn); p_pix = BlackPixel(dpy, scrn);
			} else {
				i_pix = WhitePixel(dpy, scrn); p_pix = WhitePixel(dpy, scrn);
			}
		}
	} else {
		i_pix = pixels[ink];
		p_pix = pixels[paper];
	}

/* #ifdef this just in case the compiler is braindead */
#if SCALING > 1
	x *= SCALING; y *= SCALING;
#endif

	/* Extend the bounding box if necessary */
	if (x < x_min) x_min = x;
	if (x > x_max) x_max = x + SCALING - 1;
	if (y < y_min) y_min = y;
	if (y > y_max) y_max = y + SCALING - 1;

	for (i = 0; i < 8; i++) {
		if (val & (0x80 >> i)) {
			fg_pts[nfg].x = x + i; fg_pts[nfg++].y = y;
		} else {
			bg_pts[nbg].x = x + i; bg_pts[nbg++].y = y;
		}
	}
	if (nfg) {
		if (i_pix != last_i_pix) XSetForeground(dpy, fgGc, i_pix);
		XDrawPoints(dpy, thePix, fgGc, fg_pts, nfg, CoordModeOrigin);
	}
	if (nbg) {
		if (p_pix != last_p_pix) XSetForeground(dpy, bgGc, p_pix);
		XDrawPoints(dpy, thePix, bgGc, bg_pts, nbg, CoordModeOrigin);
	}
}


/* Screen refresh functions - 3 different functions */
void 
screen_refresh_xshm(void)
{
	if (!screen_dirty) return;

	x_max |= round_up_x[SCALING - 1];

	XShmPutImage(dpy, mainWin, theGc, xim, x_min, y_min, x_min, y_min,
		x_max - (x_min - 1), y_max - (y_min - 1), False);
	XFlush(dpy);

	x_max = y_max = 0; x_min = HSize - 1; y_min = VSize - 1;
	screen_dirty = 0;
}

void 
screen_refresh_xim(void)
{
	if (!screen_dirty) return;

	x_max |= round_up_x[SCALING - 1];

	XPutImage(dpy, mainWin, theGc, xim, x_min, y_min, x_min, y_min,
		x_max - (x_min - 1), y_max - (y_min - 1));
	XFlush(dpy);

	x_max = y_max = 0; x_min = HSize - 1; y_min = VSize - 1;
	screen_dirty = 0;
}

void 
screen_refresh_pix(void)
{
	if (!screen_dirty) return;

	x_max |= round_up_x[SCALING - 1];

	XCopyArea(dpy, thePix, mainWin, theGc, 0, 0, HSize, VSize, 0, 0);
	XFlush(dpy);

	x_max = y_max = 0; x_min = HSize - 1; y_min = VSize - 1;
	screen_dirty = 0;
}


/* For 128K and +3 machines, if the screen select bit is toggled, force
 * an update of the screen to reflect the new display file.
 */
void
force_screen_refresh(void)
{
	int i;

	for (i = 0; i < 768; i++) {
		attr_write(DFILE + 6144 + i, realMem[screen_select][6144 + i]);
	}
}


/* Perform a clean close-down. */
static void
closedown(void)
{
#ifdef MITSHM
	if (GetCfg(use_mitshm)) {
		XShmDetach(dpy, &xshminfo);
		XDestroyImage(xim);
		shmdt(xshminfo.shmaddr);
		shmctl(xshminfo.shmid, IPC_RMID, 0);
	}
#endif
	XAutoRepeatOn(dpy);
	XCloseDisplay(dpy);
}


#ifdef SLOWDOWN
/* Speed up or slow down the emulator */
static void
adjust_speed(int delta)
{
	int newspd = GetCfg(slowdown) + delta;

	if (newspd < 0) newspd = GetCfg(slowdown);

	SetCfg(slowdown, newspd);
}
#endif


/* Process a KeyPress event */
static void
process_keypress(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;
	char *str;
	int mem_model;

	XLookupString(kev, buf, 2, &ks, NULL);

	/* First check for the 'special keys' */
	if (ks == key_table[KM_KEMPST_UP])
		inPorts[0x1f] |= 0x08;
	else if (ks == key_table[KM_KEMPST_DOWN])
		inPorts[0x1f] |= 0x04;
	else if (ks == key_table[KM_KEMPST_LEFT])
		inPorts[0x1f] |= 0x02;
	else if (ks == key_table[KM_KEMPST_RIGHT])
		inPorts[0x1f] |= 0x01;
	else if (ks == key_table[KM_KEMPST_FIRE])
		inPorts[0x1f] |= 0x10;
#ifdef SLOWDOWN
	else if (ks == key_table[KM_KEY_SLOWER])
		adjust_speed(1);
	else if (ks == key_table[KM_KEY_FASTER])
		adjust_speed(-1);
#endif

	switch (ks) {
		/* Emulator control keys */
		case XK_F1: case XK_F2:
			reset_cpu();
			if (GetCfg(emode) != 48) {
				last_0x7ffd = last_0x1ffd = 0;
			}
			break;
		case XK_F3:
			str = dialog_event_loop("Name of snapshot to save");
			if (str[0]) write_snapshot(str);
			break;
		case XK_F4:
			str = dialog_event_loop("Name of snapshot to load");
			if (str[0]) read_snapshot(str);
			break;
#if defined(DEV_AUDIO) || defined(PCSPKR_AUDIO) || defined(NAS_AUDIO)
		case XK_F5:	/* Toggle sound on/off */
			if (sound_on >= 0) sound_on = !sound_on;
			break;
#endif
		case XK_F7:	/* Insert new floppy disk */
			if (GetCfg(emode) == 3) {
				str =
					dialog_event_loop("New disk - enter \"[a|b]:<filename>\"");
				if (strlen(str) > 3)
					insert_disk(str + 2, *str);
			}
			break;
		case XK_F8:	/* Generate an NMI */
			if (read_byte(*pc) == 0x76)
				(*pc)++;
			PushValue(*pc);
			theProcessor.iff2 = theProcessor.iff1;
			theProcessor.iff1 = 0;
			*pc = 0x0066;
			break;
		case XK_F9:	/* switch memory model */
			str = dialog_event_loop("Memory model? (48/128/3)");
			if (!str || !str[0]) break;
			mem_model = atoi(str);
			if (mem_model == 48 || mem_model == 128 || mem_model == 3) {
				if (GetCfg(emode) != mem_model) {
					extern int basic_rom;	/* from main.c */
					SetCfg(emode, mem_model);
					basic_rom = GET_BASIC_ROM(GetCfg(emode));
					init_pages();
					read_roms();
				}
				reset_cpu();
			} else {
				xzx_mesg(XZX_ERR, "no such memory model %d", mem_model);
			}
			break;
		case XK_F10: case XK_F12:
			Quit(0);
			break;
		case XK_Tab:
			unshift_cursor = !unshift_cursor; break;

		/* Ordinary Speccy keys */
		case XK_Return:
			keyports[6] &= 0xfe; break;
		case XK_Control_L: case XK_Control_R:
		case XK_Shift_L: case XK_Shift_R:
			keyports[0] &= 0xfe; break;
		case XK_Alt_L: case XK_Alt_R:
		case XK_Meta_L: case XK_Meta_R:
			keyports[7] &= 0xfd; break;
		case XK_BackSpace: case XK_Delete:
			keyports[0] &= 0xfe; keyports[4] &= 0xfe; break;
		case XK_Escape:
			keyports[0] &= 0xfe; keyports[3] &= 0xfe; break;
		case XK_Up:
			unshift_cursor || (keyports[0] &= 0xfe);
				keyports[4] &= 0xf7; break;
		case XK_Down:
			unshift_cursor || (keyports[0] &= 0xfe);
				keyports[4] &= 0xef; break;
		case XK_Left:
			unshift_cursor || (keyports[0] &= 0xfe);
				keyports[3] &= 0xef; break;
		case XK_Right:
			unshift_cursor || (keyports[0] &= 0xfe);
				keyports[4] &= 0xfb; break;
		case XK_minus:
			keyports[7] &= 0xfd; keyports[6] &= 0xf7; break;
		case XK_underscore:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[4] &= 0xfe; break;
		case XK_equal:
			keyports[7] &= 0xfd; keyports[6] &= 0xfd; break;
		case XK_plus:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[6] &= 0xfb; break;
		case XK_semicolon:
			keyports[7] &= 0xfd; keyports[5] &= 0xfd; break;
		case XK_colon:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[0] &= 0xfd; break;
		case XK_apostrophe:
			keyports[7] &= 0xfd; keyports[4] &= 0xf7; break;
		case XK_comma:
			keyports[7] &= 0xfd; keyports[7] &= 0xf7; break;
		case XK_period:
			keyports[7] &= 0xfd; keyports[7] &= 0xfb; break;
		case XK_less:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[2] &= 0xf7; break;
		case XK_greater:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[2] &= 0xef; break;
		case XK_slash:
			keyports[7] &= 0xfd; keyports[0] &= 0xef; break;
		case XK_question:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[0] &= 0xf7; break;
		case XK_backslash: case XK_bar:
			keyports[0] = keyports[1] = keyports[2] = keyports[3] =
			keyports[4] = keyports[5] = keyports[6] = keyports[7] = 0xff;
			break;
		default:
			if (ks < 128) {
				keyports[keytable[ks].port] &= keytable[ks].bit;
			}
			break;
	}
}

/* Process a KeyRelease event */
static void
process_keyrelease(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;

	XLookupString(kev, buf, 2, &ks, NULL);

	if (ks == key_table[KM_KEMPST_UP])
		inPorts[0x1f] &= ~0x08;
	else if (ks == key_table[KM_KEMPST_DOWN])
		inPorts[0x1f] &= ~0x04;
	else if (ks == key_table[KM_KEMPST_LEFT])
		inPorts[0x1f] &= ~0x02;
	else if (ks == key_table[KM_KEMPST_RIGHT])
		inPorts[0x1f] &= ~0x01;
	else if (ks == key_table[KM_KEMPST_FIRE])
		inPorts[0x1f] &= ~0x10;

	switch (ks) {
		case XK_Return:
			keyports[6] |= 0x01; break;
		case XK_Control_L: case XK_Control_R:
		case XK_Shift_L: case XK_Shift_R:
			keyports[0] |= 0x01; break;
		case XK_Alt_L: case XK_Alt_R:
		case XK_Meta_L: case XK_Meta_R:
			keyports[7] |= 0x02; break;
		case XK_BackSpace: case XK_Delete:
			keyports[0] |= 0x01; keyports[4] |= 0x01; break;
		case XK_Escape:
			keyports[0] |= 0x01; keyports[3] |= 0x01; break;
		case XK_Up:
			unshift_cursor || (keyports[0] |= 0x01);
				keyports[4] |= 0x08; break;
		case XK_Down:
			unshift_cursor || (keyports[0] |= 0x01);
				keyports[4] |= 0x10; break;
		case XK_Left:
			unshift_cursor || (keyports[0] |= 0x01);
				keyports[3] |= 0x10; break;
		case XK_Right:
			unshift_cursor || (keyports[0] |= 0x01);
				keyports[4] |= 0x04; break;
		case XK_minus:
			keyports[7] |= 0x02; keyports[6] |= 0x08; break;
		case XK_underscore:
			keyports[0] &= 0xfe; keyports[7] |= 0x02;
			keyports[4] |= 0x01; break;
		case XK_equal:
			keyports[7] |= 0x02; keyports[6] |= 0x02; break;
		case XK_plus:
			keyports[0] &= 0xfe; keyports[7] |= 0x02;
			keyports[6] |= 0x04; break;
		case XK_semicolon:
			keyports[7] |= 0x02; keyports[5] |= 0x02; break;
		case XK_colon:
			keyports[0] &= 0xfe; keyports[7] |= 0x02;
			keyports[0] |= 0x02; break;
		case XK_apostrophe:
			keyports[7] |= 0x02; keyports[4] |= 0x08; break;
		case XK_comma:
			keyports[7] |= 0x02; keyports[7] |= 0x08; break;
		case XK_period:
			keyports[7] |= 0x02; keyports[7] |= 0x04; break;
		case XK_less:
			keyports[0] &= 0xfe; keyports[7] |= 0x02;
			keyports[2] |= 0x08; break;
		case XK_greater:
			keyports[0] &= 0xfe; keyports[7] |= 0x02;
			keyports[2] |= 0x10; break;
		case XK_slash:
			keyports[7] |= 0x02; keyports[0] |= 0x10; break;
		case XK_question:
			keyports[0] &= 0xfe; keyports[7] |= 0x02;
			keyports[0] |= 0x08; break;
		case XK_backslash: case XK_bar:
			keyports[0] = keyports[1] = keyports[2] = keyports[3] =
			keyports[4] = keyports[5] = keyports[6] = keyports[7] = 0xff;
			break;
		default:
			if (ks < 128)
				keyports[keytable[ks].port] |= ~keytable[ks].bit;
			break;
	}
}


/* Tell the window manager everything it needs to know about the 
 * top-level window (borderWin).
 */
static void
notify_wm(int *argcnt, char **argvec)
{
	Pixmap icon_pix;
	XWMHints xwmh;
	XSizeHints xsh;
	XClassHint xch;
	XTextProperty appName, iconName;
	char *app_name;
	char *icon_name="xzx";

	icon_pix = XCreatePixmapFromBitmapData(dpy,
		root, xzx_bits, xzx_width, xzx_height,
		BlackPixel(dpy, scrn), WhitePixel(dpy, scrn),
		DefaultDepth(dpy, scrn));

	app_name = (char *)malloc(20);
	if (!app_name)
		app_name = "xzx";
	else
		sprintf(app_name, "xzx %d.%d", ZX_Major, ZX_Minor);

	xsh.flags = PSize | PMinSize | PMaxSize;

	if (strlen(GetCfg(geometry)) > 0)
		xsh.flags |= USPosition;
	else
		xsh.flags |= PPosition;

	xsh.min_width  = HSize; xsh.min_height = VSize;
	xsh.max_width  = HSize + BORDER_WIDTH * 2;
	xsh.max_height = VSize + BORDER_WIDTH * 2;

	if (XStringListToTextProperty(&app_name, 1, &appName) == 0) {
		xzx_mesg(XZX_ERR, "can't create a TextProperty!");
		return;
	}
	if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
		xzx_mesg(XZX_ERR, "can't create a TextProperty!");
		return;
	}

	xwmh.initial_state = NormalState;
	xwmh.input = True;
	xwmh.icon_pixmap = icon_pix;
	xwmh.flags = StateHint | IconPixmapHint | InputHint;

	xch.res_name = "xzx";
	xch.res_class = "Xzx";

	XSetWMProperties(dpy, borderWin, &appName, &iconName, 
		argvec, *argcnt, &xsh, &xwmh, &xch);
}


/* Calculate initial window position depending on our geometry spec.
 * FIXME: this does not take into account the size of WM decorations.
 */
static void
calc_win_position(int *x_ret, int *y_ret)
{
	int x_tmp, y_tmp, w_ret, h_ret;
	int mask;

	if (!GetCfg(geometry)[0]) return;

	mask = XParseGeometry(GetCfg(geometry), &x_tmp, &y_tmp, &w_ret, &h_ret);

	*x_ret = 0; *y_ret = 0;

	if (!mask) {
		xzx_mesg(XZX_WARN, "invalid geometry specification %s", 
			GetCfg(geometry));
		GetCfg(geometry)[0] = '\0';
		return;
	}

	if (mask & XValue) 
		*x_ret = (mask & XNegative) ?
			(DisplayWidth(dpy, scrn) + x_tmp - (HSize + BORDER_WIDTH * 2)) :
			x_tmp;

	if (mask & YValue) 
		*y_ret = (mask & YNegative) ?
			(DisplayHeight(dpy, scrn) + y_tmp - (VSize + BORDER_WIDTH * 2)) :
			y_tmp;
}


/*
 * Image initialisation routines - one for Pixmaps, one for XImages
 */
static int
pixmap_init(void)
{
	thePix = XCreatePixmap(dpy, root, HSize, VSize, DefaultDepth(dpy, scrn));
	XSetForeground(dpy, fgGc, pixels[7]);	/* dull white */
	XDrawRectangle(dpy, thePix, fgGc, 0, 0, HSize, VSize);

	screen_refresh = screen_refresh_pix;
	screen_write = screen_write_pix;

	return 0;
}

static int
image_init(void)
{
#ifdef MITSHM
	if (GetCfg(use_mitshm)) {
		xim = XShmCreateImage(dpy, DefaultVisual(dpy, scrn),
			DefaultDepth(dpy, scrn), ZPixmap, NULL, &xshminfo, HSize, VSize);
		if (!xim) {
			xzx_mesg(XZX_ERR, "xzx: failed to create a shared XImage!");
			return -1;
		}
		xshminfo.shmid = shmget(IPC_PRIVATE, xim->bytes_per_line * xim->height,
			IPC_CREAT|0777);
		if (xshminfo.shmid == -1) {
			xzx_mesg(XZX_ERR, "shmget <xshminfo.shmid> failed");
			perror("  reason:");
			return -1;
		}
		xshminfo.shmaddr = xim->data = shmat(xshminfo.shmid, 0, 0);
		if (!xshminfo.shmaddr) {
			xzx_mesg(XZX_ERR, "shmat <xshminfo.shmaddr> failed");
			perror("  reason:");
			return -1;
		}
		xshminfo.readOnly = False;
		if (!XShmAttach(dpy, &xshminfo)) {
			xzx_mesg(XZX_ERR, "XShmAttach failed");
			perror("  reason:");
			return -1;
		}
		screen_refresh = screen_refresh_xshm;
	} else {
#endif
		xim = XCreateImage(dpy,  DefaultVisual(dpy, scrn),
			DefaultDepth(dpy, scrn), ZPixmap, 0, NULL, HSize, VSize, 8, 0);
		if (!xim) {
			xzx_mesg(XZX_ERR, "XCreateImage failed");
			perror("  reason:");
			return -1;
		}
		if ((xim->data = (char *)malloc(xim->bytes_per_line * xim->height))
			== NULL) {
			xzx_mesg(XZX_ERR, "malloc <xim->data> failed");
			perror("  reason:");
			return -1;
		}
		screen_refresh = screen_refresh_xim;
#ifdef MITSHM
	}
#endif
	screen_write = screen_write_xim;
	bytes_per_line = xim->bytes_per_line;
	xim_data_ptr = xim->data;

	return 0;
}


/*
 * Initialise the 16 colours required by the Speccy (or just black and white
 * if <use_mono> is set.  Return -1 if colour allocation failed.
 */
static int
get_colours(unsigned long *pix_ret, int use_mono)
{
	XColor xc;
	int i;

	if (use_mono || DefaultDepthOfScreen(sptr) <= 2) {
		for (i = 0; i < 8; i++) {
			pix_ret[i] = i < 4 ? BlackPixel(dpy,scrn) : WhitePixel(dpy,scrn);
			pix_ret[i+8] = i < 4 ? BlackPixel(dpy,scrn) : WhitePixel(dpy,scrn);
		}
		SetCfg(monochrome, True);
	} else {
		xc.flags = DoRed | DoGreen | DoBlue;
		for (i = 0; i < NCOLORS; i++) {
			xc.red   = rgb_vals[i][0];
			xc.green = rgb_vals[i][1];
			xc.blue  = rgb_vals[i][2];
			if (!XAllocColor(dpy, cmap, &xc)) {
				xzx_mesg(XZX_WARN, "XAllocColor failed on colour %d", i);
				return(-1);
			}
			pix_ret[i] = xc.pixel;
		}
	}

	return 0;
}



/**************************************************************************
 *  Public functions start here
 */

/* Start all the "hardware" up */
void
start_emulator(int *argcp, char **argvp)
{
	int i;

	key_table = (KeySym *)malloc(sizeof(KeySym) * NUM_KEYMAPPINGS);

	x_max = y_max = 0; x_min = HSize - 1; y_min = VSize - 1;

	/* Connect to a display and initialise run-time configuration */
	if ((dpy = xzx_open_display(argcp, argvp, key_table)) == NULL) {
		xzx_mesg(XZX_FATAL, "failed to open X display");
	}

	flash_phase = !GetCfg(flashing);

	/* Some useful X variables */
	scrn = DefaultScreen(dpy);
	sptr = DefaultScreenOfDisplay(dpy);
	root = DefaultRootWindow(dpy);

	theGc = XCreateGC(dpy, root, 0, NULL);
	XCopyGC(dpy, DefaultGC(dpy, scrn), ~0, theGc);
	XSetGraphicsExposures(dpy, theGc, False);

	fgGc = XCreateGC(dpy, root, 0, NULL);
	XCopyGC(dpy, DefaultGC(dpy, scrn), ~0, fgGc);
	XSetGraphicsExposures(dpy, fgGc, False);

	bgGc = XCreateGC(dpy, root, 0, NULL);
	XCopyGC(dpy, DefaultGC(dpy, scrn), ~0, bgGc);
	XSetGraphicsExposures(dpy, bgGc, False);

	/* Colour initialisation */
	want_colour = !GetCfg(monochrome);
	if (GetCfg(priv_cmap))
		cmap = XCreateColormap(dpy, root, DefaultVisual(dpy, scrn), AllocNone);
	else
		cmap  = DefaultColormap(dpy, scrn);

	if (get_colours(pixels, GetCfg(monochrome)) == -1) {
		xzx_mesg(XZX_WARN, "colour allocation failed - using private cmap");
		SetCfg(priv_cmap, True);
		cmap = XCreateColormap(dpy, root, DefaultVisual(dpy, scrn), AllocNone);
		if (get_colours(pixels, GetCfg(monochrome)) == -1)
			xzx_mesg(XZX_FATAL, "monochrome colormap initialisation failed!");
	}

	/* Get a font to use */
	theFont = XLoadQueryFont(dpy, GetCfg(font));
	if (!theFont) {
		xzx_mesg(XZX_WARN, "can't load font <%s>, trying <%s>...",
			GetCfg(font), GetCfg(font));
		theFont = XLoadQueryFont(dpy, DEF_FONT);
		if (!theFont)
			xzx_mesg(XZX_FATAL, "can't load font <%s>!", DEF_FONT);
	}
	XSetFont(dpy, theGc, theFont->fid);

	/* Image handling initialisation */
	if (GetCfg(pixmaps)) {
		if (pixmap_init() != 0)
			xzx_mesg(XZX_FATAL, "pixmap initialisation failure");
	} else if (image_init() != 0) {
#ifdef MITSHM
		xzx_mesg(XZX_WARN,
				"MIT-SHM initialisation failure, trying without MIT-SHM...");
		SetCfg(use_mitshm, False);
#endif
		if (image_init() != 0)
			xzx_mesg(XZX_FATAL, "image initialisation failure!");
#ifdef MITSHM
	}
#endif

	/* No keys initially being pressed */
	for (i = 0; i < 8; i++)
		keyports[i] = 0xff;

#if defined(PCSPKR_AUDIO) || defined(DEV_AUDIO) || defined(NAS_AUDIO)
	if (audio_init() != 0) {
		xzx_mesg(XZX_WARN, "audio initialisation failed");
		sound_on = -1;
	} else {
		sound_on = GetCfg(use_sound);
	}
#endif

#ifdef JOY
	/* Linux analogue joystick initialisation */
	if (GetCfg(use_js) && joystick_init() != 0) { 
		xzx_mesg(XZX_WARN, "joystick initialisation failed");
		SetCfg(use_js, 0);
	}
#endif

	/* This applies just as well if the actual DFILE is at 0xc000,
	 * since the top two bits are masked out anyway */
	DFILE = 0x4000;

	calc_win_position(&xpos, &ypos);

	/* Window creation */
	borderWin = XCreateSimpleWindow(dpy, root, xpos, ypos,
		HSize + BORDER_WIDTH * 2, VSize + BORDER_WIDTH * 2,
		0, pixels[7], pixels[7]);
	mainWin = XCreateSimpleWindow(dpy, borderWin, BORDER_WIDTH, BORDER_WIDTH,
		HSize, VSize, 0, pixels[7], pixels[7]); /* dull white */

	make_dialog(argcp, argvp);

	/* let's be environmentally friendly */
	notify_wm(argcp, argvp);

	if (GetCfg(priv_cmap))
		XSetWindowColormap(dpy, borderWin, cmap);
	
	XSelectInput(dpy, borderWin, KeyPressMask | KeyReleaseMask | ExposureMask |
		EnterWindowMask | LeaveWindowMask | StructureNotifyMask);
	XMapRaised(dpy, borderWin);
	XMapRaised(dpy, mainWin);

	XFlush(dpy);

	OnQuit(closedown);
}


/* Check & process any incoming X events */
void 
check_events(void)
{
	static XEvent xev;
	XConfigureEvent *conf_ev;
	XCrossingEvent *cev;

	while (XEventsQueued(dpy, QueuedAfterReading)) {
		XNextEvent(dpy, &xev);
		switch(xev.type) {
			case Expose:
				/* force a refresh of the entire screen */
				x_max = HSize - 1; y_max = VSize - 1; x_min = y_min = 0;
				screen_dirty++;
				break;
			case ConfigureNotify:
				conf_ev = (XConfigureEvent *)&xev;
				XMoveWindow(dpy, mainWin, (conf_ev->width - HSize) / 2,
				(conf_ev->height - VSize) / 2);
				break;
			case MapNotify: case UnmapNotify: case ReparentNotify:
				/* just avoid complaints about unhandled events here */
				break;
			case EnterNotify: case LeaveNotify:
				cev = (XCrossingEvent *)&xev;
				if (cev->detail != NotifyInferior)
					if (cev->type == EnterNotify)
						XAutoRepeatOn(dpy);
					else
						XAutoRepeatOff(dpy);
				break;
			case KeyPress:
				process_keypress((XKeyEvent *)&xev);
				break;
			case KeyRelease:
				process_keyrelease((XKeyEvent *)&xev);
				break;
			default:
				xzx_mesg(XZX_WARN, "unhandled X event, type %d", xev.type);
				break;
		}
	}
}


/* <val> has just been written to <addr> in the attribute area.
 * Update the screen with screen_write().
 */
void
attr_write(uns16 addr, uns8 val)
{
	int a_off = (addr & 0x3fff) - 6144;
	int d_off;

	d_off = ((a_off & 0x300) << 3) +    /* Which third of the display file */
			(a_off & 0xff) + DFILE;

	screen_write(d_off + 0x000, read_byte(d_off + 0x000));
	screen_write(d_off + 0x100, read_byte(d_off + 0x100));
	screen_write(d_off + 0x200, read_byte(d_off + 0x200));
	screen_write(d_off + 0x300, read_byte(d_off + 0x300));
	screen_write(d_off + 0x400, read_byte(d_off + 0x400));
	screen_write(d_off + 0x500, read_byte(d_off + 0x500));
	screen_write(d_off + 0x600, read_byte(d_off + 0x600));
	screen_write(d_off + 0x700, read_byte(d_off + 0x700));
}


/*
 * Input & output port emulation.
 */

/* The 128K and +3 memory management system.  We provide 2 different
 * implementations of port 0x7ffd, since it behaves slightly differently
 * between the 128K and +3 machines.
 */
void 
out_bankm_128(uns8 val)
{
	if (last_0x7ffd == val)
		return;

#ifdef DEBUG
	xzx_mesg(XZX_INFO, 
		"bankm: RAM select %d, ROM select %d, screen %d, paging %sabled",
		val & RAM_MASK, (val >> 4) & 0x01, (val >> 3) & 0x01, 
		PAGING_DISABLED(last_0x7ffd) ? "dis" : "en");
#endif

	/* First check bits D0-D2; RAM select (at 0xC000) */
	if (!PAGING_DISABLED(last_0x7ffd)) {
		page_in(3, RAM0 + (val & RAM_MASK));
		last_0x7ffd = (last_0x7ffd & ~RAM_MASK) | (val & RAM_MASK);
	}

	/* Now bit D3, screen select; 0 selects page 5, 1 selects page 7 */
	screen_select = (val & SCREEN_MASK) >> 3 ? RAM7 : RAM5;
	if ((val & SCREEN_MASK) != (last_0x7ffd & SCREEN_MASK))
		force_screen_refresh();
	last_0x7ffd = (last_0x7ffd & ~SCREEN_MASK) | (val & SCREEN_MASK);

	/* Now bit D4, ROM select (at 0x0000) */
	if (!PAGING_DISABLED(last_0x7ffd)) {
		page_in(0, ROM0 + (val >> 4) & 0x01);
		last_0x7ffd = (last_0x7ffd & ~ROM_MASK) | (val & ROM_MASK);
	}

	/* And bit D5, the page disable bit */
	if (!PAGING_DISABLED(last_0x7ffd))
		last_0x7ffd = (last_0x7ffd & ~PAGE_DISABLE) | (val & PAGE_DISABLE);
}

void 
out_bankm_pl3(uns8 val)
{
	int new_rom;

	if (last_0x7ffd == val)
		return;

#ifdef DEBUG
	xzx_mesg(XZX_INFO, 
		"bankm: RAM select %d, ROM select %d, screen %d, paging %sabled",
		val & RAM_MASK, (val >> 4) & 0x01, (val >> 3) & 0x01, 
		PAGING_DISABLED(last_0x7ffd) ? "dis" : "en");
#endif

	/* First check bits D0-D2; RAM select, could be extended */
	if (!PAGING_DISABLED(last_0x7ffd)) {
		if (!EXTENDED_RAM(last_0x1ffd))
			page_in(3, RAM0 + (val & RAM_MASK));
		last_0x7ffd = (last_0x7ffd & ~RAM_MASK) | (val & RAM_MASK);
	}

	/* Now bit D3, screen select; 0 selects page 5, 1 selects page 7 */
	screen_select = (val & SCREEN_MASK) >> 3 ? RAM7 : RAM5;
	if (val & SCREEN_MASK != last_0x7ffd & SCREEN_MASK)
		force_screen_refresh();
	last_0x7ffd = (last_0x7ffd & ~SCREEN_MASK) | (val & SCREEN_MASK);

	/* Now bit D4, ROM select (at 0x0000) */
	if (!PAGING_DISABLED(last_0x7ffd)) {
		new_rom = ((last_0x1ffd >> 1) & 0x02) | ((val >> 4) & 0x01);
		if (!EXTENDED_RAM(last_0x1ffd))
			page_in(0, new_rom);
		last_0x7ffd = (last_0x7ffd & ~ROM_MASK) | (val & ROM_MASK);
	}

	/* And bit D5, the page disable bit */
	if (!PAGING_DISABLED(last_0x7ffd))
		last_0x7ffd = (last_0x7ffd & ~PAGE_DISABLE) | (val & PAGE_DISABLE);
}

void
out_bank678(uns8 val)
{
	int new_rom;

	if (last_0x1ffd == val)
		return;

#ifdef DEBUG
	xzx_mesg(XZX_INFO, "bank678");
#endif
	
	if (!PAGING_DISABLED(last_0x1ffd)) {
		if (!EXTENDED_RAM(val)) {
			new_rom = ((last_0x7ffd >> 4) & 0x01) | ((val >> 1) & 0x02);
			page_in(0, new_rom);
			last_0x1ffd = (last_0x1ffd & ~6) | (val & 6);
			last_0x1ffd = (last_0x1ffd & ~1) | (val & 1);
#ifdef DEBUG
			xzx_mesg(XZX_INFO, "vertical ROM switch");
#endif
		} else {
#ifdef DEBUG
			xzx_mesg(XZX_INFO, "extended RAM enabled");
#endif
			switch ((val >> 1) & 0x03) {
				case 0:
					page_in(0, RAM0); page_in(1, RAM1);
					page_in(2, RAM2); page_in(3, RAM3);
					break;
				case 1:
					page_in(0, RAM4); page_in(1, RAM5);
					page_in(2, RAM6); page_in(3, RAM7);
					break;
				case 2:
					page_in(0, RAM4); page_in(1, RAM5);
					page_in(2, RAM6); page_in(3, RAM3);
					break;
				case 3:
					page_in(0, RAM4); page_in(1, RAM7);
					page_in(2, RAM6); page_in(3, RAM3);
					break;
				default:
					xzx_mesg(XZX_FATAL, "impossible switch?");
			}
			last_0x1ffd = (last_0x1ffd & ~6) | (val & 6);
			last_0x1ffd = (last_0x1ffd & ~1) | (val & 1);
		}
	}	/* if !PAGING_DISABLED ... */

#ifdef DEBUG
	if ((val & DISK_MOTOR) != (last_0x1ffd & DISK_MOTOR)) {
		if (!(val & DISK_MOTOR)) {
			xzx_mesg(XZX_INFO, "disk stopped");
		} else {
			xzx_mesg(XZX_INFO, "disk started");
		}
	}
#endif
}


void
out_byte(uns16 port, uns8 val)
{
	/* border colour and (possibly) speaker control */
	if ((port & 0xff) == 0xfe) {
		if (last_border_col != (val & 0x07)) {
			last_border_col = val & 0x07;
			XSetWindowBackground(dpy, borderWin, pixels[last_border_col]);
			XClearWindow(dpy, borderWin);
		}
#if defined(PCSPKR_AUDIO) || defined(DEV_AUDIO) || defined(NAS_AUDIO)
		/* speaker control */
		if (sound_on > 0)
			out_byte_sound(outPorts[0xfe], val);
#endif
	}

	/* banked memory management */
	if (port == 0x7ffd && GetCfg(emode) == 128)
		out_bankm_128(val);
	if (port == 0x7ffd && GetCfg(emode) == 3)
		out_bankm_pl3(val);
	if (port == 0x1ffd && GetCfg(emode) == 3)
		out_bank678(val);

#ifdef XZX_IF1
	/* Interface I microdrive I/O */
	if ((port & 0xff) == 0xef)
		out_port_ef(val);
	if ((port & 0xff) == 0xe7)
		out_port_e7(val);
#endif

#ifdef PSEUDO_IO
	if (port == PSEUDO_STDOUT)
		putc((int)val, stdout);
	else if (port == PSEUDO_STDERR)
		putc((int)val, stderr);
#endif

	outPorts[port & 0xff] = val;
}


uns8
in_byte(uns16 port)
{
	int res = 0xff;
	int highbyte;

	/* kempston joystick port */
	if ((port & 0xff) == 0x1f) {
#ifdef JOY
		if (GetCfg(use_js)) inPorts[0x1f] = joystick_read();
#endif
		return inPorts[0x1f];
	} else if ((port & 0xff) == 0xfe) { /* keyboard */
		highbyte = port >> 8;
		if (!(highbyte & 0x80)) res &= keyports[7];
		if (!(highbyte & 0x40)) res &= keyports[6];
		if (!(highbyte & 0x20)) res &= keyports[5];
		if (!(highbyte & 0x10)) res &= keyports[4];
		if (!(highbyte & 0x08)) res &= keyports[3];
		if (!(highbyte & 0x04)) res &= keyports[2];
		if (!(highbyte & 0x02)) res &= keyports[1];
		if (!(highbyte & 0x01)) res &= keyports[0];
		return res;
	} else
#ifdef XZX_IF1
		if ((port & 0xff) == 0xe7)
			return in_port_e7();
		else if ((port & 0xff) == 0xef)
			return in_port_ef();
		else
#endif
#ifdef PSEUDO_IO
			if (port == PSEUDO_STDIN)
				return (uns8)getchar();
			else
#endif
			
				return 0xff;
}


/* Any byte in the attribute area with the FLASH bit set has
 * its ink and paper swapped twice/sec.
 * FIXME: this routine is somewhat inefficient at the moment.
 */
void
do_flashing(void)
{
	int i;
	uns8 val;

	if (!GetCfg(flashing))
		return;

	flash_phase = 1 - flash_phase;

	for (i = 0; i < 768; i++) {
		if ((val = realMem[screen_select][6144 + i]) & 0x80)
			attr_write((DFILE + 6144) + i, val);
	}
}



/*******************************************
 * Popup dialog handling routines start here
 *
 */

/* Create the popup dialog box */
static void
make_dialog(int *argcnt, char **argvec)
{
	XWMHints xwmh;
	XSizeHints xsh;
	XSetWindowAttributes xswa;
	XTextProperty dlgName;
	char *dlg_name;

	/* first work out the necessary geometries */
	font_ht = theFont->ascent + theFont->descent;
	/* only works with fixed width fonts! */
	char_width = XTextWidth(theFont, " ", 1);

	label_wid = char_width * EF_MAX_VISIBLE;
	label_ht = font_ht;
	entry_wid = (char_width * EF_MAX_VISIBLE) + 4;
	entry_ht = font_ht + 4;
	ok_wid = (char_width * strlen(BUTTON_LABEL)) + 4;
	ok_ht = font_ht + 4;
	dlg_wid = entry_wid + (2 * GUTTER_WID);
	dlg_ht = label_ht + entry_ht + ok_ht + (4 * GUTTER_HT);

	/* window creation */
	dlg_win = XCreateSimpleWindow(dpy, root, dlg_x, dlg_y, dlg_wid, dlg_ht, 
		0, pixels[0], pixels[7]);	/* black & dull white */
	entry_win = XCreateSimpleWindow(dpy, dlg_win, GUTTER_WID, 
		(2 * GUTTER_HT) + label_ht, entry_wid, entry_ht, 0,
		pixels[0], pixels[7]);
	ok_win = XCreateSimpleWindow(dpy, dlg_win, GUTTER_WID, 
		(3 * GUTTER_HT) + label_ht + entry_ht, ok_wid, ok_ht, 0,
		pixels[0], pixels[7]);


	/* tell the window manager */
	xsh.flags = PSize | PMinSize | PMaxSize | PPosition;
	xsh.min_width  = dlg_wid; xsh.min_height = dlg_ht;
	xsh.max_width  = dlg_wid; xsh.max_height = dlg_ht;

	xwmh.initial_state = NormalState;
	xwmh.input = True;
	xwmh.flags = StateHint | InputHint;

	dlg_name = "xzx popup";
	if (XStringListToTextProperty(&dlg_name, 1, &dlgName) == 0) {
		xzx_mesg(XZX_ERR, "can't create a TextProperty!");
		return;
	}

	xswa.save_under = True;

	if (GetCfg(priv_cmap))
		XSetWindowColormap(dpy, dlg_win, cmap);
	
	XSelectInput(dpy, dlg_win, ExposureMask | KeyPressMask | EnterWindowMask);
	XSelectInput(dpy, ok_win, ButtonPressMask | ButtonReleaseMask);

	XChangeWindowAttributes(dpy, dlg_win, CWSaveUnder, &xswa);
	XSetTransientForHint(dpy, dlg_win, borderWin);
	XSetWMProperties(dpy, dlg_win, &dlgName, NULL,
		argvec, *argcnt, &xsh, &xwmh, NULL);

	/* subwindows must be mapped & raised */
	XMapRaised(dpy, entry_win);
	XMapRaised(dpy, ok_win);
}


static void
paint_entry_field()
{
	int real_curpos;

	/* draw the string */
	XSetForeground(dpy, theGc, pixels[0]);
	XSetBackground(dpy, theGc, pixels[7]);
	XDrawImageString(dpy, entry_win, theGc, 2, 2 + theFont->ascent, 
		dlg_string + cur_win_pos, strlen(dlg_string + cur_win_pos));

	/* draw the cursor */
	real_curpos = (cursor_pos - cur_win_pos) * char_width + 2;
	XFillRectangle(dpy, entry_win, theGc,
		real_curpos, 2, char_width, entry_ht - 4);

	/* blank the rest of the window */
	XClearArea(dpy, entry_win, real_curpos + char_width, 2,
		(EF_MAX_VISIBLE * char_width) - (real_curpos + char_width + 2),
		font_ht, False);
}


static void
paint_dialog(char *label)
{
	/* draw entry field and button borders */
	XSetForeground(dpy, theGc, pixels[0]);
	XDrawLine(dpy, entry_win, theGc, 0, entry_ht - 1, 0, 0);
	XDrawLine(dpy, entry_win, theGc, 0, 0, entry_wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[15]);
	XDrawLine(dpy, entry_win, theGc, entry_wid - 1, 0,
		entry_wid - 1, entry_ht - 1);
	XDrawLine(dpy, entry_win, theGc, entry_wid - 1, entry_ht - 1,
		0, entry_ht - 1);

	XSetForeground(dpy, theGc, pixels[15]);
	XDrawLine(dpy, ok_win, theGc, 0, ok_ht - 1, 0, 0);
	XDrawLine(dpy, ok_win, theGc, 0, 0, ok_wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[0]);
	XDrawLine(dpy, ok_win, theGc, ok_wid - 1, 0, ok_wid - 1, ok_ht - 1);
	XDrawLine(dpy, ok_win, theGc, ok_wid - 1, ok_ht - 1, 0, ok_ht - 1);

	/* paint the label & the button label */
	XDrawString(dpy, dlg_win, theGc, GUTTER_WID, GUTTER_HT + theFont->ascent,
		label, strlen(label));
	XDrawString(dpy, ok_win, theGc, 2, 2 + theFont->ascent,
		BUTTON_LABEL, strlen(BUTTON_LABEL));

	paint_entry_field();
}


static void
shadow_in(Window w, int wid, int ht)
{
	XSetForeground(dpy, theGc, pixels[0]);
	XDrawLine(dpy, w, theGc, 0, ht - 1, 0, 0);
	XDrawLine(dpy, w, theGc, 0, 0, wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[15]);
	XDrawLine(dpy, w, theGc, wid - 1, 0, wid - 1, ht - 1);
	XDrawLine(dpy, w, theGc, wid - 1, ht - 1, 0, ht - 1);
}

static void
shadow_out(Window w, int wid, int ht)
{
	XSetForeground(dpy, theGc, pixels[15]);
	XDrawLine(dpy, w, theGc, 0, ht - 1, 0, 0);
	XDrawLine(dpy, w, theGc, 0, 0, wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[0]);
	XDrawLine(dpy, w, theGc, wid - 1, 0, wid - 1, ht - 1);
	XDrawLine(dpy, w, theGc, wid - 1, ht - 1, 0, ht - 1);
}


/* Process a keypress in the dialog entry field.  Return a value of 1 if
 * we're finished typing text, 0 otherwise.
 */
static int
process_ef_key(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;

	XLookupString(kev, buf, 2, &ks, NULL);

	switch (ks) {
		case XK_Return:
			return 1;
			break;
		case XK_Escape:
			dlg_string[0] = '\0';
			return 1;
			break;
		case XK_BackSpace: case XK_Delete:
			if (cursor_pos > 0) cursor_pos--;
			dlg_string[cursor_pos] = '\0';
			if (cursor_pos < cur_win_pos) cur_win_pos--;
			break;
		default:
			if (ks < 128 && ks > 31) {
				/* add a character */
				dlg_string[cursor_pos] = (char) ks;
				if (cursor_pos < EF_MAX - 1) cursor_pos++;
				if (cursor_pos >= cur_win_pos + EF_MAX_VISIBLE) cur_win_pos++;
				dlg_string[cursor_pos] = '\0';
			}
			break;
	}

	paint_entry_field();
	XFlush(dpy);

	return 0;
}


/* Pop up the dialog box, and process events until Return is pressed or
 * the OK button is clicked.  Then pop down the dialog box and return the
 * string in the entry field.
 */
char *
dialog_event_loop(char *label)
{
	int finished = 0;
	XEvent xev;
	Window r_ret, w_ret;
	int dx, dy, winx, winy, mask_ret; 

	/* reset everything */
	dlg_string[0] = '\0'; cursor_pos = 0; cur_win_pos = 0;

	/* erase the old label */
	XClearArea(dpy, dlg_win, GUTTER_WID, GUTTER_HT, label_wid, label_ht, False);

	/* move the dialog window to under the pointer */
	if (!XQueryPointer(dpy, root, &r_ret, &w_ret, &dx, &dy,
		&winx, &winy, &mask_ret)) {
		dx = 0; dy = 0;
	} else {
		dx -= dlg_wid / 2; dy -= dlg_ht / 2;
	}
	if (dx < 0) dx = 0;
	if (dy < 0) dy = 0;
	XMoveWindow(dpy, dlg_win, dx, dy);

	XMapRaised(dpy, dlg_win);

	while (!finished) {
		XNextEvent(dpy, &xev);
		switch(xev.type) {
			case Expose:
				if (xev.xexpose.window == mainWin ||
					xev.xexpose.window == borderWin) {
					x_max = HSize - 1; y_max = VSize - 1; x_min = y_min = 0;
					screen_dirty = 1;
					screen_refresh();
				} else {
					paint_dialog(label);
				}
				break;
			case ButtonPress:
				if (xev.xbutton.window == ok_win) {
					shadow_in(xev.xbutton.window, ok_wid, ok_ht);
				}
				break;
			case ButtonRelease:
				if (xev.xbutton.window == ok_win) {
					shadow_out(xev.xbutton.window, ok_wid, ok_ht);
					finished = 1;
				}
				break;
			case KeyPress:
				if (xev.xkey.window == entry_win ||
					xev.xkey.window == dlg_win) {
					finished = process_ef_key((XKeyEvent *)&xev);
				}
				break;
			case EnterNotify:
				if (xev.xcrossing.window == dlg_win)
					XAutoRepeatOn(dpy);
				break;
			default:	/* ignore any other events */
				break;
					
		}
	}

	XUnmapWindow(dpy, dlg_win);
	XFlush(dpy);

	return dlg_string;
}
